#!perl
#
# This is MakeWeb.
# Copyright (C) 1996-2001 by Hiroshi Yuki <hyuki@hyuki.com>
# http://www.hyuki.com/makeweb/
#
# This program is free software; you can redistribute it and/or
# modify it under the same terms as Perl itself.
#
=pod

=head1 NAME

MakeWeb - A Text-to-HTML Converter.

=head1 SYNOPSIS

    perl makeweb.pl input.txt output.html

=head1 INPUT EXAMPLE

    *Section One
    -(
    Hello!
    -)
    *Section Two
    -[
    - Item
    - Item
    - Item
    -]

=head1 OUTPUT EXAMPLE

    <h2 class="section"><a name="i1" id="i1">Section One</a></h2>
    <p>Hello!</p>
    <h2 class="section"><a name="i2" id="i2">Section Two</a></h2>
    <ul>
        <li>Item</li>
        <li>Item</li>
        <li>Item</li>
    </ul>

=cut

$VERSION = "4.0.0 Alpha 5";
use strict "vars";
use Getopt::Std;

# Extention to target type.
my %ext_to_type = (
    "s?html?" => "html",
    "tex" => "latex",
);

# Command line options
my %options;
getopts('hdv', \%options);

# $debug = 1 for debug, = 0 for non-debug.
my $debug = 0;
if ($options{d}) {
    $debug = 1;
}

# $verbose = 1 for verbose information.
my $verbose = 0;
if ($options{v}) {
    $verbose = 1;
}

if ($options{h}) {
    &report_usage();
}

my $expand_level = 0;

# File name and line number stack for error messages (cf. do_infile).
my @fileinfo_stack = ();

# Current level of <ul>, <ol>, <li>, <dl>, <dt>, and <dd>.
my @li_stack = ();

# Command definitions.
# These are 'hook commands'.
my %define = (
    'hr'        => '<hr />',        # ----
    'ulbegin'   => '<ul>',          # -[
    'ulend'     => '</ul>',         # -]
    'olbegin'   => '<ol>',          # -<
    'olend'     => '</ol>',         # ->
    'libegin'   => '<li>',          # - item
    'liend'     => '</li>',         #
    'dlbegin'   => '<dl>',          # =[
    'dlend'     => '</dl>',         # =]
    'dtbegin'   => '<dt>',          # = word
    'dtend'     => '</dt>',         #
    'ddbegin'   => '<dd>',          # description
    'ddend'     => '</dd>',         #
    'embegin'   => '<em>',          # !line or <<string>>
    'emend'     => '</em>',         #
    'strongbegin'   => '<strong>',  # <*strong*>
    'strongend'     => '</strong>', #
    'varbegin'  => '<var>',         # <|var_name|>
    'varend'    => '</var>',        #
    'kbdbegin'  => '<kbd>',         # <%key sequence%>
    'kbdend'    => '</kbd>',        #
    'codebegin' => '<code>',        # <{code fragment}>
    'codeend'   => '</code>',       #
    'filenamebegin' => '<code class="filename">',   # <"filename">
    'filenameend'   => '</code>',                   #
    'pbegin'    => '<p>',           # -(
    'pmid'      => '</p><p>',       # --
    'pend'      => '</p>',          # -)
    'prebegin'  => '<pre>',         # --( and ---(
    'preend'    => '</pre>',        # --) and ---)
    'h1begin'   => '<h2 class="section">',              # *Heading
    'h1end'     => '</h2>',                             #
    'h2begin'   => '<h3 class="subsection">',           # -*Heading
    'h2end'     => '</h3>',                             #
    'h3begin'   => '<h3 class="subsubsection">',        # --*Heading
    'h3end'     => '</h3>',                             #
    'display_divbegin'  => '<div class="display">',     # =<
    'display_divend'    => '</div>',                    # =>
    'display_pbegin'        => '<div class="display"><p>',  # =(
    'display_pend'          => '</p></div>',                # =)
    'display_prebegin'      => '<pre class="display">', # ==( and ===(
    'display_preend'        => '</pre>',                # ==) and ===)
    'blockquote_pbegin'     => '<blockquote><p>',       # -((
    'blockquote_pend'       => '</p></blockquote>',     # -))
    'blockquote_prebegin'   => '<blockquote><pre>',     # --(( and ---((
    'blockquote_preend'     => '</pre></blockquote>',   # --)) and ---))
    'display_blockquote_pbegin'     => '<blockquote class="display"><p>',       # =((
    'display_blockquote_pend'       => '</p></blockquote>',                     # =))
    'display_blockquote_prebegin'   => '<blockquote class="display"><pre>',     # ==(( and ===((
    'display_blockquote_preend'     => '</pre></blockquote>',                   # ==)) and ===))
);

# ID for contents (cf. do_contents).
my $current_id = 1;

# Output format.
my $target_type = "html";

&main;

exit(0);

# Main.
sub main {
    &verbose_print("Start of MakeWeb.\n");

    &report_usage() if (@ARGV < 2);
    my $input_filename = $ARGV[0];
    my $output_filename = $ARGV[1];

    &verbose_print("Input file is $input_filename.\n");
    &verbose_print("Output file is $output_filename.\n");

    # What is the output format?
    foreach my $ext (sort keys %ext_to_type) {
        if ($output_filename =~ /\.${ext}$/) {
            $target_type = $ext_to_type{$ext};
            &verbose_print("Target type is $target_type.\n");
            last;
        }
    }

    my $output = &do_infile($input_filename);
    &do_outfile($output_filename, $output);

    &verbose_print("End of MakeWeb.\n");
}

# Is the format HTML?
sub is_html {
    return $target_type eq "html";
}

# Is the format LaTeX?
sub is_latex {
    return $target_type eq "latex";
}

# Read $infile and process all the lines.
sub do_infile {
    my ($infile) = @_;

    &verbose_print("Reading from $infile.\n");

    my (@input);
    if (!open(INPUT, $infile)) {
        &report_error("I cannot open $infile to read.\n");
        exit(1);
    }
    push(@fileinfo_stack, { filename => $infile, linenumber => 0 });
    @input = <INPUT>;
    close(INPUT);
    my $out = &do_lines($fileinfo_stack[-1], @input);
    pop(@fileinfo_stack);
    return $out;
}

# Write $str to $outfile.
sub do_outfile {
    my ($outfile, $str) = @_;

    &verbose_print("Writing to $outfile.\n");

    if (!open(OUTPUT, "> $outfile")) {
        &report_error("I cannot open $outfile to write.\n");
        exit(1);
    }
    print OUTPUT $str;
    close(OUTPUT);
}

=head1 GENERAL COMMANDS

=head2 COMMENT

    ;comment string (; must be placed on the first column)

=head2 VERBATIM

    |verbatim string
    -|string containing <, >, ", &, ...

=head2 EMPHASIS

    !strong

=head2 FILE INCLUSION

    -include=filename

=head2 SECTION HEADING

    MakeWeb     Hook commands               HTML

    *string     <-h1begin>string<-h1end>    <h2 class="section">string</h2>
    -*string    <-h2begin>string<-h2end>    <h3 class="subsection">string</h3>
    --*string   <-h3begin>string<-h3end>    <h3 class="subsubsection">string</h3>

=head2 CONTENTS COMMANDS

    -contents
    -detailcontents

=head2 TIME

    -localtime

=head2 DEBUG

    -print_define

=head1 LIST COMMANDS

=head2 UNORDERED LIST COMMANDS

    MakeWeb         Hook commands               HTML

    -[              <-ulbegin>                  <ul>
    - item          <-libegin>item<-liend>      <li>item</li>
    - item          <-libegin>item<-liend>      <li>item</li>
    - item          <-libegin>item<-liend>      <li>item</li>
    -]              <-ulend>                    </ul>

=head2 ORDERED LIST COMMANDS

    MakeWeb         Hook commands               HTML

    -<              <-olbegin>                  <ol>
    - item          <-libegin>item<-liend>      <li>item</li>
    - item          <-libegin>item<-liend>      <li>item</li>
    - item          <-libegin>item<-liend>      <li>item</li>
    ->              <-olend>                    </ol>

=head2 DESCRIPTION LIST COMMANDS

    MakeWeb         Hook commands                   HTML

    =[              <-dlbegin>                      <dl>
    = item          <-dtbegin>item<-dtend>          <dt>item</dt>
    description     <-ddbegin>description<-ddend>   <dd>description</dd>
    = item          <-dtbegin>item<-dtend>          <dt>item</dt>
    description     <-ddbegin>description<-ddend>   <dd>description</dd>
    = item          <-dtbegin>item<-dtend>          <dt>item</dt>
    description     <-ddbegin>description<-ddend>   <dd>description</dd>
    =]              <-dlend>                        </dl>

=head1 PARAGRAPH COMMANDS

     -( ... -)         <-pbegin> ... <-pend>
     --( ... --)       <-prebegin> ... <-preend>
     ---( ... ---)     <-prebegin> ... <-preend> with escape
    
     =( ... =)         <-display_pbegin> ... <-display_pend>
     ==( ... ==)       <-display_prebegin> ... <-display_preend>
     ===( ... ===)      <-display_prebegin> ... <-display_preend> with escape
    
     -(( ... -))       <-blockquote_pbegin> ... <-blockquote_pend>
     --(( ... --))     <-blockquote_prebegin> ... <-blockquote_preend>
     ---(( ... ---))   <-blockquote_prebegin> ... <-blockquote_preend> with escape
    
     =(( ... =))       <-display_blockquote_pbegin> ... <-display_blockquote_pend>
     ==(( ... ==))     <-display_blockquote_prebegin> ... <-display_blockquote_preend>
     ===(( ... ===))   <-display_blockquote_prebegin> ... <-display_blockquote_preend> with escape
    
     -mycommand( ... -mycommand)       <-mycommand_pbegin> ... <-mycommand_pend>
     --mycommand( ... --mycommand)     <-mycommand_prebegin> ... <-mycommand_preend>
     ---mycommand( ... ---mycommand)   <-mycommand_prebegin> ... <-mycommand_preend> with escape
    
     =mycommand( ... =mycommand)       <-display_mycommand_pbegin> ... <-display_mycommand_pend>
     ==mycommand( ... ==mycommand)     <-display_mycommand_prebegin> ... <-display_mycommand_preend>
     ===mycommand( ... ===mycommand)   <-display_mycommand_prebegin> ... <-display_mycommand_preend> with escape

     =< ... =>          <-display_divbegin> ... <-display_divend>
     =mycommand< ... =mycommand>    <-mycommand_divbegin> ... <-mycommand_divend>

=head1 COMMAND DEFINITION

=head2 ONE-LINE

    -mycommand=some string

=head2 MULTI-LINE

    -begin=mycommand
    first line
    second line
    third line
    -end=mycommand

=head1 CONDITIONAL COMMANDS

-if ... -endif cannot nest.

    -if=html
    ...
    -endif=html

    -if=latex
    ...
    -endif=latex

=cut

# Expand all the lines.
sub do_lines {
    my ($fileinforef, @input) = @_;
    my @orginput = @input;
    my $if_type = undef;
    my $out = '';
    while (@input) {
        $_ = next_line(\@input, $fileinforef);
        if (/^;/) {
            # Skip MakeWeb comments.
        } elsif (/^\|(.*)/) {
            # Verbatim.
            $out .= $1 . "\n";
        } elsif (/^-\|(.*)/) {
            # Verbatim but escape.
            $out .= &escape($1) . "\n";
        } elsif (/^!(.*)/) {
            # Strong.
            $out .= &expand_it("<<" . &expand_it($1) . ">>\n");
        } elsif (/^-include=(.*)/) {
            # Include file.
            $out .=  &do_infile(&expand_it($1));
            # &print_define;
        } elsif (/^-if=(\w+)/) {
            # Conditional begin.
            my $type = $1;
            if ($if_type) {
                # Cannot nest.
                &report_error("-if cannot nest.\n");
            } elsif ($type eq $target_type) {
                # Save the current type.
                $if_type = $type;
            } else {
                # Skip until -endif=xxxx.
                while (@input) {
                    local $_ = next_line(\@input, $fileinforef);
                    if (/^-endif=(\w+)/) {
                        if ($1 ne $type) {
                            &report_error("Mismatch -if=$type .. -endif=$1.\n");
                        }
                        last;
                    }
                }
            }
        } elsif (/^-endif=(\w+)/) {
            # Conditional end.
            my $type = $1;
            if (not $if_type) {
                &report_error("Missing -if=$type.");
            } else {
                # Now out of the -if ... -endif.
                $if_type = undef;
            }
        } elsif (/^-begin=(\w+)/) {
            # Define a command with multiple lines.
            my $command = $1;
            my $def = '';
            my $defined = 0;
        DEFINE:
            while (@input) {
                local $_ = next_line(\@input, $fileinforef);
                if (/^-end=(\w+)/) {
                    if ($command ne $1) {
                        &report_error("Mismatch -begin=$command ... -end=$1.\n");
                    }
                    $define{$command} = $def;
                    &debug_print("expand_line:$command is defined as $def");
                    $defined = 1;
                    last DEFINE;
                } elsif (!&can_define($_)) {
                    &report_error("The following line cannot appear within a command definition.\n$_");
                } else {
                    $def .= $_;
                }
            }
            if (!$defined) {
                &report_error("Missing -end=$1.\n");
            }
        } elsif (/^-end=(\w+)/) {
            # Invalid -end.
            &report_error("Missing -begin=$1.\n");
        } elsif (/^-(\w+)=(.*)/) {
            # -greeting=Hello, world.
            $define{$1} = $2;
        } elsif (/^-((detail)?contents)$/) {
            # Contents.
            # Future: /^-(detail|list|figure)?contents$/
            if (exists($define{$1})) {
                $out .= &expand_it($define{$1});
            } else {
                $out .= &do_contents($2, \@orginput);
            }
        } elsif (/^-localtime$/) {
            # Localtime.
            $out .= "" . localtime;
        } elsif (/^-print_define$/) {
            # Print %define for debug.
            $out .= &get_define;
        } elsif (/^-(\w+)(\n)?$/) {
            # Expand command.
            my $command = $1;
            my $newline = $2;
            $out .= &do_lines(undef, &make_lines($define{$command}));
            $out .= "\n" if ($newline);
        } elsif (/^(-{0,2})\*(\[([^[]+?)\])?(.+)/) { # Heading pattern
            # Heading.
            # *Greeting         (level 1)
            # -*Greeting        (level 2)
            # --*Greeting       (level 3)
            # *[hello]Greeting  (level 1 with id "hello")
            #
            # Note: The inside string of [  ] is used for ID.
            # It must be unique in a document.
            #
            my $level = length($1) + 1;
            my $id = $3;
            my $title = $4;
            if ($id eq '') {
                $id = "i$current_id";
                $current_id++;
            }
            my $begintag = sprintf("<-h%d%s>", $level, "begin");
            my $endtag = sprintf("<-h%d%s>", $level, "end");
            $out .= &expand_it($begintag);

            # Selfmark
            if (&is_html) {
                my $selfmark = &expand_it("<-selfmark$level>");
                $out .= qq|<a href="#$id" title="$id">$selfmark</a>| if ($selfmark);
            }

            $out .= qq|<a name="$id" id="$id">| if (&is_html);
            $out .= &expand_it($title);
            $out .= qq|</a>| if (&is_html);
            
            $out .= &expand_it($endtag);
            $out .= qq|\n|;
        } elsif (/^>(.+)/) {
            # Link.
            #
            # >My Home Page
            # http://www.hyuki.com/
            #
            # -[
            # >- My Home Page
            # http://www.hyuki.com/
            # -]
            #
            # -[
            # >- My Home Page
            # http://www.hyuki.com/
            # -[
            # - Another Item
            # -]
            # -]
            #
            my $caption = $1;
            chomp(my $url = next_line(\@input, $fileinforef));
            &debug_print("\$caption = $caption, \$url = $url\n");
            $url = &expand_it($url);
            if ($caption =~ /^- (.+)/) {
                # Special treat for -[ and -]
                if ($li_stack[-1] eq "li") {
                    $out .= &expand_it("<-liend>\n");
                    pop(@li_stack);
                }
                $caption = $1;
                $out .= &expand_it("<-libegin>\n");
                $out .= qq|<a href="$url">| if (&is_html);
                $out .= &expand_it($caption);
                $out .= qq|</a>| if (&is_html);
                $out .= "\n";
                # Do NOT output "li" element, but push "li" here.
                push(@li_stack, "li");
            } else {
                $out .= qq|<a href="$url">| if (&is_html);
                $out .= &expand_it($caption);
                $out .= qq|</a>| if (&is_html);
            }
        } elsif (/^- (.+)$/) {
            # Special treat for -[ and -]
            if ($li_stack[-1] eq "li") {
                $out .= &expand_it("<-liend>\n");
                pop(@li_stack);
            }
            $out .= &expand_it("<-libegin>\n");
            $out .= &expand_it($1);
            $out .= "\n";
            push(@li_stack, "li");
        } elsif (/^= (.*)/) {
            # Special treat for =[ and =]
            if ($li_stack[-1] eq "dd") {
                $out .= &expand_it("<-ddend>\n");
                pop(@li_stack);
            }
            $out .= &expand_it("<-dtbegin>$1<-dtend>\n");
            $out .= &expand_it("<-ddbegin>\n");
            push(@li_stack, "dd");
        } elsif (/^-(\[|\<|\]|\>)$/) {
            # -[
            # - item
            # - item
            # -]
            #
            # -<
            # - item
            # - item
            # ->
            #
            my %info = (
                "-["    => { element => "-ulbegin", type => "ul", },
                "-]"    => { element => "-ulend", type => "ul", },
                "-<"    => { element => "-olbegin", type=> "ol", },
                "->"    => { element => "-olend", type=> "ol", },
            );
            my $rawcmd = $&;
            if (not exists($info{$rawcmd})) {
                &report_error("Unknown command $rawcmd.\n");
            }
            my $elem = $info{$rawcmd}->{element};
            my $type = $info{$rawcmd}->{type};
            if ($elem =~ /begin$/) {
                push(@li_stack, $type);
            } else {
                if ($li_stack[-1] eq "li") {
                    $out .= &expand_it("<-liend>");
                    pop(@li_stack);
                }
                my $begintype = pop(@li_stack);  # error check?
                if ($begintype ne $type) {
                    &report_error("Mismatch $rawcmd.\n");
                }
            }
            $out .= &expand_it("<$elem>");
        } elsif (/^(-{1,3}|={1,3})(\w*)(\){1,2})$/) {
            my $bar = $1;   # '-', '--', '---', '=', '==', '==='
            my $cmd = $2;   # 'mycommand'
            my $par = $3;   # '(', '(('
            my $ren = "(" x length($par);
            &report_error("Missing $bar$cmd$ren.\n");
        } elsif (/^(-{1,3}|={1,3})(\w*)(\({1,2})$/) {
            # Paragraph commands.
            my %elem = (
                "-("    => "pbegin",
                "-)"    => "pend",
                "--("   => "prebegin",
                "--)"   => "preend",
                "---("  => "prebegin",
                "---)"  => "preend",
                "=("    => "display_pbegin",
                "=)"    => "display_pend",
                "==("   => "display_prebegin",
                "==)"   => "display_preend",
                "===("  => "display_prebegin",
                "===)"  => "display_preend",
                "-(("   => "blockquote_pbegin",
                "-))"   => "blockquote_pend",
                "--(("  => "blockquote_prebegin",
                "--))"  => "blockquote_preend",
                "---((" => "blockquote_prebegin",
                "---))" => "blockquote_preend",
                "=(("   => "display_blockquote_pbegin",
                "=))"   => "display_blockquote_pend",
                "==(("  => "display_blockquote_prebegin",
                "==))"  => "display_blockquote_preend",
                "===((" => "display_blockquote_prebegin",
                "===))" => "display_blockquote_preend",
            );
            my $bar = $1;   # '-', '--', '---', '=', '==', '==='
            my $cmd = $2;   # 'mycommand'
            my $par = $3;   # '(', '(('
            my $ren;
            if ($par =~ m|\(|) {
                $ren = ")" x length($par);
            } else {
                $ren = "(" x length($par);
            }
            my $escape = (length($bar) == 3 ? 1 : 0);
            my $beginline = "$bar$cmd$par";
            my $endline = "$bar$cmd$ren";
            my $begincmd;
            my $endcmd;
            if ($cmd) {
                my $display = ($bar =~ /^=/ ? "display_" : "");
                my $pre = (length($bar) == 1 ? "p" : "pre");
                $begincmd = "$display${cmd}_${pre}begin";
                $endcmd = "$display${cmd}_${pre}end";
                if (not exists($define{$begincmd})) {
                    &report_error("-$begincmd is not defined.\n");
                }
                if (not exists($define{$endcmd})) {
                    &report_error("-$endcmd is not defined.\n");
                }
            } else {
                $begincmd = $elem{$bar.$par};
                $endcmd = $elem{$bar.$ren};
            }
            my $begintag = "<-$begincmd>";
            my $endtag = "<-$endcmd>";
            my $closed = 0;
            $out .= &expand_it($begintag);
            if ($escape) {
                while (@input) {
                    $_ = next_line(\@input, $fileinforef);
                    if (/^\Q$beginline\E$/) {
                        &report_error("Missing $endline.\n");
                    } elsif (/^\Q$endline\E$/) {
                        $closed = 1;
                        last;
                    }
                    $out .= &escape($_);
                }
            } else {
                my @in;
                while (@input) {
                    $_ = next_line(\@input, $fileinforef);
                    if (/^\Q$beginline\E$/) {
                        &report_error("Missing $endline.\n");
                    } elsif (/^\Q$endline\E$/) {
                        $closed = 1;
                        last;
                    } elsif (/^--$/) {
                        push(@in, "<-pend><-pbegin>");
                        next;
                    } else {
                        push(@in, $_);
                    }
                }
                $out .= &do_lines(undef, @in);
            }
            unless ($closed) {
                &report_error("Missing $endline.\n");
            }
            $out .= &expand_it($endtag);
            $out .= "\n";
        } elsif (/^=(\w*)(\<|\>)$/) {
            # Division command.
            my $class = ($1 eq "" ? "display" : $1);
            my $beginend = ($2 eq "<" ? "begin" : "end");
            my $elem = "${class}_div${beginend}";
            if (not exists($define{$elem})) {
                &report_error("-$elem is not defined (for $&).\n");
            }
            $out .= &expand_it("<-$elem>\n");
        } elsif (/^=(\[|\])$/) {
            #
            # =[
            # = item
            # description
            # = item
            # description
            # =]
            #
            my $rawcmd = $&;
            my $beginend = ($1 eq "[" ? "begin" : "end");
            my $elem = "dl${beginend}";
            if (not exists($define{$elem})) {
                &report_error("-$elem is not defined (for $rawcmd).\n");
            }
            if ($beginend eq "begin") {
                push(@li_stack, "dl");
            } else {
                if ($li_stack[-1] eq "dd") {
                    $out .= &expand_it("<-ddend>\n");
                    pop(@li_stack);
                }
                my $begintype = pop(@li_stack);
                if ($begintype ne "dl") {
                    &report_error("Mismatch $rawcmd.\n");
                }
            }
            $out .= &expand_it("<-$elem>\n");
        } elsif (/^----$/) {
            # Horizontal rule.
            $out .= &expand_it("<-hr>");
        } elsif (0) {
            # Your new command will be added here.
        } else {
            # Non-command.
            $out .= &expand_it($_);
        }
    }
    # End of lines. Begin to verify.
    if ($if_type) {
        &report_error("Missing -endif=$if_type.\n");
    }
    return $out;
}

=head1 INLINE COMMANDS

    MakeWeb             Escape  Hook commands                   HTML
    <-command>          -       -                               (Expand -command)
    <<emphasis>>        no      <-embegin> ... <-emend>         <em> ... </em>
    <*strong*>          no      <-strongbegin> ... <-strongend> <strong> ... </strong>
    <|var_name|>        yes     <-varbegin> ... <-varend>       <var> ... </var>
    <%key sequence%>    yes     <-kbdbegin> ... <-kbdend>       <kbd> ... </kbd>
    <{code fragment}>   yes     <-codebegin> ... <-codeend>     <code> ... </code>
    <"filename">        yes     <-filenamebegin> ... <-filenameend>     <code class="filename"> ... </code>
    <-myclass<string>>  no      <-myclass_spanbegin> ... <-myclass_spaneend>    <span class="myclass"> ... </span>

=cut

# Expand inline command.
sub expand_it {
    local $_ = shift @_;

    $expand_level++;
    if ($expand_level > 100) {
        &report_error("Recursive definition in $_.\n");
    }

    # Expand <-command>
    s[ \<-(\w+?)\> ][ &do_lines(undef, &make_lines($define{$1})) ]gex;

    # Expand << ... >>
    s[ \<\<(.+?)\>\> ][ &expand_it("<-embegin>$1<-emend>") ]gex;

    # Expand <* ... *>
    s[ \<\*(.+?)\*\> ][ &expand_it("<-strongbegin>$1<-strongend>") ]gex;

    # Expand <| ... |>
    s[ \<\|(.+?)\|\> ][ &expand_it("<-varbegin>") . &escape($1) . &expand_it("<-varend>") ]gex;

    # Expand <% ... %>
    s[ \<\%(.+?)\%\> ][ &expand_it("<-kbdbegin>") . &escape($1) . &expand_it("<-kbdend>") ]gex;

    # Expand <{ ... }>
    s[ \<\{(.+?)\}\> ][ &expand_it("<-codebegin>") . &escape($1) . &expand_it("<-codeend>") ]gex;

    # Expand <" ... ">
    s[ \<\"(.+?)\"\> ][ &expand_it("<-filenamebegin>") . &escape($1) . &expand_it("<-filenameend>") ]gex;

    # Expand <-myclass< ... >>
    s[ \<-(\w+)\<(.+?)\>\> ][ &expand_span($1, $2) ]gex;

    &debug_print("&expand_it returns = |$_|\n");

    $expand_level--;

    return $_;
}

# Make <-myclass_spanbegin>...<-myclass_spanend>
# That will be <span class="myclass">...</span>.
sub expand_span {
    my ($class, $content) = @_;
    my $begin = "${class}_spanbegin";
    my $end = "${class}_spanend";
    if (not exists($define{$begin})) {
        &report_error("-$begin is not defined.\n");
    }
    if (not exists($define{$end})) {
        &report_error("-$end is not defined.\n");
    }
    return &expand_it("<-$begin>$content<-$end>");
}

# Report error and exit.
sub report_error {
    my ($msg) = @_;
    my $prefix;
    if (@fileinfo_stack) {
        $prefix = $fileinfo_stack[-1]->{filename} . "(" . $fileinfo_stack[-1]->{linenumber} . "):";
    }
    print "$prefix$msg";
    exit(1);
}

# Report the usage and exit.
sub report_usage {
    print "This is MakeWeb $main::VERSION.\n";
    print "http://www.hyuki.com/makeweb/\n";
    print "\n";
    print "Usage: perl makeweb.pl [options] input.txt output.html\n";
    print "Example 1 to make convert sample.txt to sample.html:\n";
    print "    perl makeweb.pl sample.txt sample.html\n";
    print "Example 2 to show help:\n";
    print "    perl makeweb.pl -h\n";
    print "Options:\n";
    print "    -d   shows debug information.\n";
    print "    -h   shows this usage.\n";
    print "    -v   shows verbose information.\n";
    print "\n";
    exit(1);
}

# Debug print.
sub debug_print {
    my ($msg) = @_;
    if ($debug) {
        print $msg;
    }
}

# Can the string appear in -begin= ... -end= ... ?
sub can_define {
    my ($s) = @_;
    if ($s =~ /^-begin=\w+/) {
        return 0;
    } elsif ($s =~ /^-end=\w+/) {
        return 0;
    } else {
        return 1;
    }
}

# Make contents.
sub do_contents {
    my ($detail, $inref) = @_;
    my $curlevel = 1;
    my $curid = 1;
    my $result .= qq|-[\n|;
    foreach (@$inref) {
        if (/^(-{0,2})\*(\[([^[]+?)\])?(.+)/) {  # Heading pattern
            my ($level, $id, $title) = (length($1) + 1, $3, $4);
            if ($id eq '') {
                $id = "i$curid";
                $curid++;
            }
            next if ($level > 1 and not $detail);
            if ($curlevel < $level) {
                $result .= qq|-[\n|;
            } elsif ($level < $curlevel) {
                while ($level < $curlevel){
                    $result .= "-]\n";
                    $curlevel--;
                }
            }
            $result .= qq|>- $title\n|;
            $result .= qq|#$id\n|;
            $curlevel = $level;
        }
    }
    while ($curlevel > 1) {
        $result .= "-]\n";
        $curlevel--;
    }
    $result .= "-]\n";
    return &do_lines(undef, &make_lines($result));
}

# Split a string into array without removing \n.
sub make_lines {
    local $_ = shift @_;
    my (@arr);
    while (/(.*\n?)/g) {
        push(@arr, $1);
    }
    return @arr;
}

# < => &lt;, 
sub escape {
    local $_ = shift;
    if (&is_html) {
        s/&/&amp;/g;
        s/</&lt;/g;
        s/>/&gt;/g;
        s/"/&quot;/g;
        # s/'/&apos;/g;
    } elsif (&is_latex) {
        s/\\/\\\\/g;
        s/#/\\#/g;
        s/_/\\_/g;
        s/\$/\\\$/g;
        s/\{/\\{/g;
        s/\}/\\}/g;
    }
    $_;
}

# Shift next line.
sub next_line {
    my ($inputlistref, $fileinforef) = @_;
    my $line = shift @$inputlistref;
    if ($fileinforef) {
        $fileinforef->{linenumber}++;
        &debug_print($fileinforef->{filename} . "(" . $fileinforef->{linenumber} . ")\n");
        &debug_print($line);
    }
    return $line;
}

# Current definitions.
sub get_define {
    my $out;
    foreach (sort keys %define) {
        $out .= $_ . "=>" . $define{$_} . "\n";
    }
    return $out;
}

# Print verbose information.
sub verbose_print {
    my ($msg) = @_;
    if ($verbose) {
        print $msg;
    }
}

1;
__END__

=head1 AUTHOR

Copyright (C) 1996-2001 Hiroshi Yuki.

hyuki@hyuki.com

This program is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=head1 SEE ALSO

=over 4

=item MakeWeb Home Page

http://www.hyuki.com/makeweb/

=back

=head1 TODO

    - LaTeX output.
        - Entity reference (< to &lt;) conversion is not valid in LaTeX.
    - Plain text output.
    - -contents must exist in only outmost level.
    - doc: -contents is for file.
    - doc: Expansion timing and Version 4.0.
    - doc: <span>
    - doc: <<...>>
    - doc: <|...|>
    - doc: <%...%>
    - doc: <{...}>
    - doc: <"...">
    - doc: <-myclass< ... >>
    - ID uniqness check.
    - <img>
    - <table> ...
    - 'title' attribute in 'a' element.
    - 'alt', 'width', 'height', attribute in 'img' element.
    - 'style' attribute?

=cut
